from typing import ClassVar

from pylint.checkers import BaseChecker

WARNING_CODE = "W1701"


class DunderDelChecker(BaseChecker):
    """info: This class is used by pylint that checks
    if "__del__" is not used
    if "__del__" is used then a warning will be raised.
    """

    name = "dunder-del"
    priority = -1
    msgs: ClassVar = {
        WARNING_CODE: (
            '"__del__" is not allowed!',
            "no-dunder-del",
            '"__del__" is not allowed! A "with" block could be a good solution',
        ),
    }

    def visit_functiondef(self, node):
        """:param node: info about a function or method
        :return: None
        """
        if node.name == "__del__":
            self.add_message(WARNING_CODE, node=node)


def register(linter):
    """Required method to auto register this checker."""
    linter.register_checker(DunderDelChecker(linter))



"""
Pylint Smartup Transformers
===========================

This plugin will register some transform functions which will allow PyLint to better
understand some classed used in Salt which trigger, `no-member` and `maybe-no-member`
A bridge between the `pep8`_ library and PyLint

"""

from astroid import MANAGER
from astroid import nodes


def rootlogger_transform(obj):
    if obj.name != "RootLogger":
        return

    def _inject_method(cls, msg, *args, **kwargs):
        pass

    if not hasattr(obj, "trace"):
        obj.trace = _inject_method

    if not hasattr(obj, "garbage"):
        obj.garbage = _inject_method


def register(linter):
    """Register the transformation functions."""
    try:
        MANAGER.register_transform(nodes.Class, rootlogger_transform)
    except AttributeError:
        MANAGER.register_transform(nodes.ClassDef, rootlogger_transform)



"""
saltpylint.thirdparty
~~~~~~~~~~~~~~~~~~~~~

Checks all imports against a list of known and allowed 3rd-party modules
and raises a lint error if an import not in that known 3rd-party modules list
is not gated.
"""

import os
import pkgutil
from typing import ClassVar

import astroid
import astroid.exceptions
from astroid.modutils import is_relative
from astroid.modutils import is_standard_module
from pylint.checkers import BaseChecker

MSGS = {
    "W8410": (
        "3rd-party module import is not gated in a try/except: %r",
        "3rd-party-module-not-gated",
        "3rd-party module imported without being gated in a try/except.",
    ),
    "C8410": (
        "3rd-party local module import is not gated in a try/except: %r. "
        "Consider importing at the module global scope level and gate it "
        "in a try/except.",
        "3rd-party-local-module-not-gated",
        "3rd-party module locally imported without being gated. Consider importing "
        "at the module global scope level and gate it in a try/except",
    ),
}


def get_import_package(modname):
    """Return the import package.

    Given modname is 'salt.utils', returns 'salt'
    """
    return modname.split(".")[0]


class ThirdPartyImportsChecker(BaseChecker):
    name = "3rd-party-imports"
    msgs = MSGS
    priority = -2

    options = (
        (
            "allowed-3rd-party-modules",
            {
                "default": (),
                "type": "csv",
                "metavar": "<3rd-party-modules>",
                "help": "Known 3rd-party modules which don' require being gated, separated by a comma",
            },
        ),
    )

    known_py3_modules: ClassVar = ["builtins"]

    unix_modules = (
        "posix",
        "pwd",
        "spwd",
        "grp",
        "crypt",
        "termios",
        "tty",
        "pty",
        "fcntl",
        "pipes",
        "resource",
        "nis",
        "syslog",
        "posixpath",
    )
    win_modules = ("msilib", "msvcrt", "winreg", "winsound", "ntpath")

    all_modules: ClassVar = {m[1]: m[0] for m in pkgutil.iter_modules()}
    std_modules_path: ClassVar = all_modules["os"]
    std_modules: ClassVar = []
    for mod, path in all_modules.items():
        if path == std_modules_path and mod not in unix_modules + win_modules:
            std_modules.append(mod)

    known_std_modules = known_py3_modules + std_modules

    def __init__(self, linter=None) -> None:
        BaseChecker.__init__(self, linter)
        self._inside_try_except = False
        self._inside_funcdef = False
        self._inside_if = False
        self.cwd = None
        self.allowed_3rd_party_modules = []

    def open(self):
        super().open()
        self.cwd = os.getcwd()
        self.allowed_3rd_party_modules = set(
            self.linter.config.allowed_3rd_party_modules,
        )  # pylint: disable=no-member

    # pylint: disable=unused-argument
    def visit_if(self, node):
        self._inside_if = True

    def leave_if(self, node):
        self._inside_if = True

    def visit_try(self, node):
        self._inside_try_except = True

    def leave_try(self, node):
        self._inside_try_except = False

    def visit_functiondef(self, node):
        self._inside_funcdef = True

    def leave_functiondef(self, node):
        self._inside_funcdef = False

    # pylint: enable=unused-argument

    def visit_import(self, node):
        names = [name for name, _ in node.names]
        for name in names:
            self._check_third_party_import(node, name)

    def visit_importfrom(self, node):
        self._check_third_party_import(node, node.modname)

    def _check_third_party_import(self, node, modname):
        if modname in self.known_std_modules:
            # Don't even care about these
            return
        module_file = node.root().file

        if is_relative(modname, module_file):
            # Is the import relative to the curent module being checked
            return

        base_modname = modname.split(".", 1)[0]
        import_modname = modname
        while True:
            try:
                imported_module = node.do_import_module(import_modname)
                if not imported_module:
                    break
                if imported_module.file.startswith(self.cwd):
                    # This is an import to package under the project being tested
                    return
                # If we reached this far, we were able to import the module but it's
                # not considered a module from within the project being checked
                break
            except Exception:  # pylint: disable=broad-except
                # This is, for example, from salt.ext.six.moves import Y
                # Because `moves` is a dynamic/runtime module
                import_modname = import_modname.rsplit(".", 1)[0]

            if import_modname == base_modname or not import_modname:
                break

        try:
            if not is_standard_module(modname) and self._inside_try_except is False:
                if get_import_package(modname) in self.allowed_3rd_party_modules:
                    return
                if self._inside_if or self._inside_funcdef:
                    message_id = "3rd-party-local-module-not-gated"
                else:
                    message_id = "3rd-party-module-not-gated"
                self.add_message(message_id, node=node, args=modname)
        except astroid.exceptions.AstroidBuildingException:
            # Failed to import
            if self._inside_try_except is False:
                if get_import_package(modname) in self.allowed_3rd_party_modules:
                    return
                if self._inside_if or self._inside_funcdef:
                    message_id = "3rd-party-local-module-not-gated"
                else:
                    message_id = "3rd-party-module-not-gated"
                self.add_message(message_id, node=node, args=modname)
        except astroid.exceptions.InferenceError:
            # Failed to import
            if self._inside_try_except is False:
                if get_import_package(modname) in self.allowed_3rd_party_modules:
                    return
                if self._inside_if or self._inside_funcdef:
                    message_id = "3rd-party-local-module-not-gated"
                else:
                    message_id = "3rd-party-module-not-gated"
                self.add_message(message_id, node=node, args=modname)
        except ImportError:
            # Definitly not a standard library import
            if self._inside_try_except is False:
                if get_import_package(modname) in self.allowed_3rd_party_modules:
                    return
                if self._inside_if or self._inside_funcdef:
                    message_id = "3rd-party-local-module-not-gated"
                else:
                    message_id = "3rd-party-module-not-gated"
                self.add_message(message_id, node=node, args=modname)


def register(linter):
    """Required method to auto register this checker."""
    linter.register_checker(ThirdPartyImportsChecker(linter))



"""
PyLint File Permissions Check Plugin
====================================

PyLint plugin which checks for specific file permissions
"""

import glob
import os
import stat
import sys
from typing import ClassVar

from pylint.checkers import BaseChecker


class FilePermsChecker(BaseChecker):
    """Check for files with undesirable permissions."""

    name = "fileperms"
    msgs: ClassVar = {
        "E0599": (
            "Module file has the wrong file permissions(expected %s): %s",
            "file-perms",
            ("Wrong file permissions"),
        ),
    }

    priority = -1

    options = (
        (
            "fileperms-default",
            {
                "default": "0644",
                "type": "string",
                "metavar": "ZERO_PADDED_PERM",
                "help": "Desired file permissons. Default: 0644",
            },
        ),
        (
            "fileperms-ignore-paths",
            {
                "default": (),
                "type": "csv",
                "metavar": "<comma-separated-list>",
                "help": "File paths to ignore file permission. Glob patterns allowed.",
            },
        ),
    )

    def process_module(self, node):
        """Process a module."""
        for listing in self.config.fileperms_ignore_paths:
            if node.file.split(f"{os.getcwd()}/")[-1] in glob.glob(listing):
                # File is ignored, no checking should be done
                return

        desired_perm = self.config.fileperms_default
        if "-" in desired_perm:
            desired_perm = desired_perm.split("-")
        else:
            desired_perm = [desired_perm]

        if len(desired_perm) > 2:  # noqa: PLR2004
            msg = "Permission ranges should be like XXXX-YYYY"
            raise RuntimeError(msg)

        for idx, _perm in enumerate(desired_perm):
            desired_perm[idx] = desired_perm[idx].strip('"').strip("'").lstrip("0").zfill(4)
            if desired_perm[idx][0] != "0":
                # Always include a leading zero
                desired_perm[idx] = f"0{desired_perm[idx]}"
            if desired_perm[idx][1] != "o":
                desired_perm[idx] = "0o" + desired_perm[idx][1:]
            if sys.platform.startswith("win"):
                # Windows does not distinguish between user/group/other.
                # They must all be the same. Also, Windows will automatically
                # set the execution bit on files with a known extension
                # (eg .exe, .bat, .com). So we cannot reliably test the
                # execution bit on other files such as .py files.
                user_perm_noexec = int(desired_perm[idx][-3])
                if user_perm_noexec % 2 == 1:
                    user_perm_noexec -= 1
                desired_perm[idx] = desired_perm[idx][:-3] + (str(user_perm_noexec) * 3)

        module_perms = oct(stat.S_IMODE(os.stat(node.file).st_mode))

        if len(desired_perm) == 1:
            if module_perms != desired_perm[0]:
                if sys.platform.startswith("win"):
                    # Check the variant with execution bit set due to the
                    # unreliability of checking the execution bit on Windows.
                    user_perm_noexec = int(desired_perm[0][-3])
                    desired_perm_exec = desired_perm[0][:-3] + (str(user_perm_noexec + 1) * 3)
                    if module_perms == desired_perm_exec:
                        return
                self.add_message("E0599", line=1, args=(desired_perm[0], module_perms))
        elif module_perms < desired_perm[0] or module_perms > desired_perm[1]:
            if sys.platform.startswith("win"):
                # Check the variant with execution bit set due to the
                # unreliability of checking the execution bit on Windows.
                user_perm_noexec0 = int(desired_perm[0][-3])
                desired_perm_exec0 = desired_perm[0][:-3] + (str(user_perm_noexec0 + 1) * 3)
                user_perm_noexec1 = int(desired_perm[1][-3])
                desired_perm_exec1 = desired_perm[1][:-3] + (str(user_perm_noexec1 + 1) * 3)
                if desired_perm_exec0 <= module_perms <= desired_perm_exec1:
                    return
            desired_perm = ">= {} OR <= {}".format(*desired_perm)
            self.add_message("E0599", line=1, args=(desired_perm, module_perms))


def register(linter):
    """Required method to auto register this checker."""
    linter.register_checker(FilePermsChecker(linter))



from typing import ClassVar

import astroid
from pylint.checkers import BaseChecker

VIRT_LOG = "log-in-virtual"


class VirtChecker(BaseChecker):
    """checks for compliance inside __virtual__."""

    name = "virt-checker"

    msgs: ClassVar = {
        "E1401": (
            "Log statement detected inside __virtual__ function. Remove it.",
            VIRT_LOG,
            "Loader processes __virtual__ so logging not in scope",
        ),
    }
    options = ()

    priority = -1

    def visit_functiondef(self, node):
        """Verifies no logger statements inside __virtual__."""
        if (
            not isinstance(node, astroid.FunctionDef)
            or node.is_method()
            or node.type != "function"
            or not node.body
        ):
            # only process functions
            return

        try:
            if node.name != "__virtual__":
                # only need to process the __virtual__ function
                return
        except AttributeError:
            return

        # walk contents of __virtual__ function
        for child in node.get_children():
            for functions in child.get_children():
                if isinstance(functions, astroid.Call) and isinstance(
                    functions.func,
                    astroid.Attribute,
                ):
                    try:
                        # Inspect each statement for an instance of 'logging'
                        for inferred in functions.func.expr.infer():
                            try:
                                instance_type = inferred.pytype().split(".")[0]
                            except TypeError:
                                continue
                            if instance_type == "logging":
                                self.add_message(VIRT_LOG, node=functions)
                                # Found logger, don't need to keep processing this line
                                break
                    except AttributeError:
                        # Not a log function
                        return


def register(linter):
    """Required method to auto register this checker."""
    linter.register_checker(VirtChecker(linter))



"""
saltpylint.blacklist
~~~~~~~~~~~~~~~~~~~~

Checks blacklisted imports and code usage on salt
"""

import contextlib
import fnmatch
import os

import astroid
from pylint.checkers import BaseChecker
from pylint.checkers import utils

BLACKLISTED_IMPORTS_MSGS = {
    "E9402": (
        "Uses of a blacklisted module %r: %s",
        "blacklisted-module",
        "Used a module marked as blacklisted is imported.",
    ),
    "E9403": (
        "Uses of a blacklisted external module %r: %s",
        "blacklisted-external-module",
        "Used a module marked as blacklisted is imported.",
    ),
    "E9404": (
        "Uses of a blacklisted import %r: %s",
        "blacklisted-import",
        "Used an import marked as blacklisted.",
    ),
    "E9405": (
        "Uses of an external blacklisted import %r: %s",
        "blacklisted-external-import",
        "Used an external import marked as blacklisted.",
    ),
    "E9406": (
        "Uses of blacklisted test module execution code: %s",
        "blacklisted-test-module-execution",
        "Uses of blacklisted test module execution code.",
    ),
    "E9407": (
        "Uses of blacklisted sys.path updating through 'ensure_in_syspath'. "
        "Please remove the import and any calls to 'ensure_in_syspath()'.",
        "blacklisted-syspath-update",
        "Uses of blacklisted sys.path updating through ensure_in_syspath.",
    ),
}


class BlacklistedImportsChecker(BaseChecker):
    name = "blacklisted-imports"
    msgs = BLACKLISTED_IMPORTS_MSGS
    priority = -2

    def open(self):
        self.blacklisted_modules = (
            "salttesting",
            "integration",
            "unit",
            "mock",
            "six",
            "distutils.version",
            "unittest",
            "unittest2",
        )

    def visit_import(self, node):
        """Triggered when an import statement is seen."""
        module_filename = node.root().file
        if fnmatch.fnmatch(module_filename, "__init__.py*") and not fnmatch.fnmatch(
            module_filename,
            "test_*.py*",
        ):
            return
        node.root()
        names = [name for name, _ in node.names]

        for name in names:
            self._check_blacklisted_module(node, name)

    def visit_importfrom(self, node):
        """Triggered when a from statement is seen."""
        module_filename = node.root().file
        if fnmatch.fnmatch(module_filename, "__init__.py*") and not fnmatch.fnmatch(
            module_filename,
            "test_*.py*",
        ):
            return
        basename = node.modname
        self._check_blacklisted_module(node, basename)

    def _check_blacklisted_module(self, node, mod_path):
        """Check if the module is blacklisted."""
        for mod_name in self.blacklisted_modules:
            if mod_path == mod_name or mod_path.startswith(mod_name + "."):
                names = []
                for name, name_as in node.names:
                    if name_as:
                        names.append(f"{name} as {name_as}")
                    else:
                        names.append(name)
                try:
                    import_from_module = node.modname
                    if import_from_module == "salttesting.helpers":
                        for name in names:
                            if name == "ensure_in_syspath":
                                self.add_message("blacklisted-syspath-update", node=node)
                                continue
                            msg = f"Please use 'from tests.support.helpers import {name}'"
                            self.add_message("blacklisted-module", node=node, args=(mod_path, msg))
                        continue
                    if import_from_module in (
                        "salttesting.mock",
                        "mock",
                        "unittest.mock",
                        "unittest2.mock",
                    ):
                        for name in names:
                            msg = f"Please use 'from tests.support.mock import {name}'"
                            if import_from_module in (
                                "salttesting.mock",
                                "unittest.mock",
                                "unittest2.mock",
                            ):
                                message_id = "blacklisted-module"
                            else:
                                message_id = "blacklisted-external-module"
                            self.add_message(message_id, node=node, args=(mod_path, msg))
                        continue
                    if import_from_module == "salttesting.parser":
                        for name in names:
                            msg = f"Please use 'from tests.support.parser import {name}'"
                            self.add_message("blacklisted-module", node=node, args=(mod_path, msg))
                        continue
                    if import_from_module == "salttesting.case":
                        for name in names:
                            msg = f"Please use 'from tests.support.case import {name}'"
                            self.add_message("blacklisted-module", node=node, args=(mod_path, msg))
                        continue
                    if import_from_module == "salttesting.unit":
                        for name in names:
                            msg = f"Please use 'from tests.support.unit import {name}'"
                            self.add_message("blacklisted-module", node=node, args=(mod_path, msg))
                        continue
                    if import_from_module.startswith(("unittest", "unittest2")):
                        for name in names:
                            msg = f"Please use 'from tests.support.unit import {name}'"
                            self.add_message("blacklisted-module", node=node, args=(mod_path, msg))
                        continue
                    if import_from_module == "salttesting.mixins":
                        for name in names:
                            msg = f"Please use 'from tests.support.mixins import {name}'"
                            self.add_message("blacklisted-module", node=node, args=(mod_path, msg))
                        continue
                    if import_from_module == "six":
                        for name in names:
                            msg = f"Please use 'from salt.ext.six import {name}'"
                            self.add_message("blacklisted-module", node=node, args=(mod_path, msg))
                        continue
                    if import_from_module == "distutils.version":
                        for name in names:
                            msg = f"Please use 'from salt.utils.versions import {name}'"
                            self.add_message("blacklisted-module", node=node, args=(mod_path, msg))
                        continue
                    if names:
                        for name in names:
                            if name in (
                                "TestLoader",
                                "TextTestRunner",
                                "TestCase",
                                "expectedFailure",
                                "TestSuite",
                                "skipIf",
                                "TestResult",
                            ):
                                msg = f"Please use 'from tests.support.unit import {name}'"
                                self.add_message(
                                    "blacklisted-module",
                                    node=node,
                                    args=(mod_path, msg),
                                )
                                continue
                            if name in ("SaltReturnAssertsMixin", "SaltMinionEventAssertsMixin"):
                                msg = f"Please use 'from tests.support.mixins import {name}'"
                                self.add_message(
                                    "blacklisted-module",
                                    node=node,
                                    args=(mod_path, msg),
                                )
                                continue
                            if name in ("ModuleCase", "SyndicCase", "ShellCase", "SSHCase"):
                                msg = f"Please use 'from tests.support.case import {name}'"
                                self.add_message(
                                    "blacklisted-module",
                                    node=node,
                                    args=(mod_path, msg),
                                )
                                continue
                            if name == "run_tests":
                                self.add_message(
                                    "blacklisted-test-module-execution",
                                    node=node,
                                    args=(
                                        "Please remove the 'if __name__ == \"__main__\":' section from the "
                                        "end of the module"
                                    ),
                                )
                                continue
                            if mod_name in ("integration", "unit"):
                                if name in (
                                    "SYS_TMP_DIR",
                                    "TMP",
                                    "FILES",
                                    "PYEXEC",
                                    "MOCKBIN",
                                    "SCRIPT_DIR",
                                    "TMP_STATE_TREE",
                                    "TMP_PRODENV_STATE_TREE",
                                    "TMP_CONF_DIR",
                                    "TMP_SUB_MINION_CONF_DIR",
                                    "TMP_SYNDIC_MINION_CONF_DIR",
                                    "TMP_SYNDIC_MASTER_CONF_DIR",
                                    "CODE_DIR",
                                    "TESTS_DIR",
                                    "CONF_DIR",
                                    "PILLAR_DIR",
                                    "TMP_SCRIPT_DIR",
                                    "ENGINES_DIR",
                                    "LOG_HANDLERS_DIR",
                                    "INTEGRATION_TEST_DIR",
                                ):
                                    msg = f"Please use 'from tests.support.paths import {name}'"
                                    self.add_message(
                                        "blacklisted-import",
                                        node=node,
                                        args=(mod_path, msg),
                                    )
                                    continue
                                self.add_message(
                                    "blacklisted-import",
                                    node=node,
                                    args=(
                                        mod_path,
                                        f"Please use 'from tests.{mod_path} import {name}'",
                                    ),
                                )
                                continue
                            self.add_message(
                                "blacklisted-module",
                                node=node,
                                args=(
                                    mod_path,
                                    "Please report this error to SaltStack so we can fix it: "
                                    f"Trying to import {name} from {mod_path}",
                                ),
                            )
                except AttributeError:
                    if mod_name in (
                        "integration",
                        "unit",
                        "mock",
                        "six",
                        "distutils.version",
                        "unittest",
                        "unittest2",
                    ):
                        if mod_name in ("integration", "unit"):
                            msg = f"Please use 'import tests.{mod_name} as {mod_name}'"
                            message_id = "blacklisted-import"
                        elif mod_name == "mock":
                            msg = f"Please use 'import tests.support.{mod_name} as {mod_name}'"
                            message_id = "blacklisted-external-import"
                        elif mod_name == "six":
                            msg = f"Please use 'import salt.ext.{name} as {name}'"
                            message_id = "blacklisted-external-import"
                        elif mod_name == "distutils.version":
                            msg = "Please use 'import salt.utils.versions' instead"
                            message_id = "blacklisted-import"
                        elif mod_name.startswith(("unittest", "unittest2")):
                            msg = f"Please use 'import tests.support.unit as {mod_name}' instead"
                            message_id = "blacklisted-import"
                        self.add_message(message_id, node=node, args=(mod_path, msg))
                        continue
                    msg = f"Please report this error to SaltStack so we can fix it: Trying to import {mod_path}"
                    self.add_message("blacklisted-import", node=node, args=(mod_path, msg))


BLACKLISTED_LOADER_USAGE_MSGS = {
    "E9501": (
        "Blacklisted salt loader dunder usage. Setting dunder attribute %r to module %r. "
        "Use 'salt.support.mock' and 'patch.dict()' instead.",
        "unmocked-patch-dunder",
        "Uses a blacklisted salt loader dunder usage in tests.",
    ),
    "E9502": (
        "Blacklisted salt loader dunder usage. Setting attribute %r to module %r. "
        "Use 'salt.support.mock' and 'patch()' instead.",
        "unmocked-patch",
        "Uses a blacklisted salt loader dunder usage in tests.",
    ),
    "E9503": (
        "Blacklisted salt loader dunder usage. Updating dunder attribute %r on module %r. "
        "Use 'salt.support.mock' and 'patch.dict()' instead.",
        "unmocked-patch-dunder-update",
        "Uses a blacklisted salt loader dunder usage in tests.",
    ),
}


class BlacklistedLoaderModulesUsageChecker(BaseChecker):
    name = "blacklisted-unmocked-patching"
    msgs = BLACKLISTED_LOADER_USAGE_MSGS
    priority = -2

    def open(self):
        self.process_module = False
        self.salt_dunders = (
            "__opts__",
            "__salt__",
            "__runner__",
            "__context__",
            "__utils__",
            "__ext_pillar__",
            "__thorium__",
            "__states__",
            "__serializers__",
            "__ret__",
            "__grains__",
            "__pillar__",
            "__sdb__",
            "__proxy__",
            "__low__",
            "__orchestration_jid__",
            "__running__",
            "__intance_id__",
            "__lowstate__",
            "__env__",
        )
        self.imported_salt_modules = {}

    def close(self):
        self.process_module = False
        self.imported_salt_modules = {}

    def visit_module(self, node):
        module_filename = node.root().file
        if not fnmatch.fnmatch(os.path.basename(module_filename), "test_*.py*"):
            return
        self.process_module = True

    def leave_module(self, node):
        if self.process_module:
            # Reset
            self.process_module = False
            self.imported_salt_modules = {}

    def visit_import(self, node):
        """Triggered when an import statement is seen."""
        if self.process_module:
            # Store salt imported modules
            for module, import_as in node.names:
                if not module.startswith("salt"):
                    continue
                if import_as and import_as not in self.imported_salt_modules:
                    self.imported_salt_modules[import_as] = module
                    continue
                if module not in self.imported_salt_modules:
                    self.imported_salt_modules[module] = module

    def visit_importfrom(self, node):
        """Triggered when a from statement is seen."""
        if self.process_module:
            if not node.modname.startswith("salt"):
                return
            # Store salt imported modules
            for module, import_as in node.names:
                if import_as and import_as not in self.imported_salt_modules:
                    self.imported_salt_modules[import_as] = import_as
                    continue
                if module not in self.imported_salt_modules:
                    self.imported_salt_modules[module] = module

    def visit_assign(self, node, *args):
        if not self.process_module:
            return

        node_left = node.targets[0]

        if isinstance(node_left, astroid.Subscript):
            # Were're changing an existing attribute
            if not isinstance(node_left.value, astroid.Attribute):
                return
            if node_left.value.attrname in self.salt_dunders:
                if node_left.value.expr.name in self.imported_salt_modules:
                    self.add_message(
                        "unmocked-patch-dunder-update",
                        node=node,
                        args=(
                            node_left.value.attrname,
                            self.imported_salt_modules[node_left.value.expr.name],
                        ),
                    )
                return

        if not isinstance(node_left, astroid.AssignAttr):
            return

        try:
            if node_left.expr.name not in self.imported_salt_modules:
                # If attributes are not being set on salt's modules,
                # leave it alone, for now!
                return
        except AttributeError:
            # This might not be what we're looking for
            return

        # we're assigning to an imported salt module!
        if node_left.attrname in self.salt_dunders:
            # We're changing salt dunders
            self.add_message(
                "unmocked-patch-dunder",
                node=node,
                args=(node_left.attrname, self.imported_salt_modules[node_left.expr.name]),
            )
            return

        # Changing random attributes
        self.add_message(
            "unmocked-patch",
            node=node,
            args=(node_left.attrname, self.imported_salt_modules[node_left.expr.name]),
        )


RESOURCE_LEAKAGE_MSGS = {
    "W8470": ("Resource leakage detected. %s ", "resource-leakage", "Resource leakage detected."),
}


class ResourceLeakageChecker(BaseChecker):
    name = "resource-leakage"
    msgs = RESOURCE_LEAKAGE_MSGS
    priority = -2

    def open(self):
        self.inside_with_ctx = False

    def close(self):
        self.inside_with_ctx = False

    def visit_with(self, node):
        self.inside_with_ctx = True

    def leave_with(self, node):
        self.inside_with_ctx = False

    def visit_call(self, node):
        if isinstance(node.func, astroid.Attribute):
            if node.func.attrname == "fopen" and self.inside_with_ctx is False:
                msg = (
                    "Please call 'salt.utils.files.fopen' using the 'with' context "
                    "manager, otherwise the file handle won't be closed and "
                    "resource leakage will occur."
                )
                self.add_message("resource-leakage", node=node, args=(msg,))
        elif (
            isinstance(node.func, astroid.Name)
            and utils.is_builtin(node.func.name)
            and node.func.name == "open"
        ):
            if self.inside_with_ctx:
                msg = (
                    "Please use 'with salt.utils.files.fopen()' instead of "
                    "'with open()'. It assures salt does not leak "
                    "file handles."
                )
            else:
                msg = (
                    "Please use 'salt.utils.files.fopen()' instead of 'open()' "
                    "using the 'with' context manager, otherwise the file "
                    "handle won't be closed and resource leakage will occur."
                )
            self.add_message("resource-leakage", node=node, args=(msg,))


MOVED_TEST_CASE_CLASSES_MSGS = {
    "E9490": (
        "Moved test case base class detected. %s",
        "moved-test-case-class",
        "Moved test case base class detected.",
    ),
    "E9491": (
        "Moved test case mixin class detected. %s",
        "moved-test-case-mixin",
        "Moved test case mixin class detected.",
    ),
}


class MovedTestCaseClassChecker(BaseChecker):
    name = "moved-test-case-class"
    msgs = MOVED_TEST_CASE_CLASSES_MSGS
    priority = -2

    def open(self):
        self.process_module = False

    def close(self):
        self.process_module = False

    def visit_module(self, node):
        module_filename = node.root().file
        if not fnmatch.fnmatch(os.path.basename(module_filename), "test_*.py*"):
            return
        self.process_module = True

    def leave_module(self, node):
        if self.process_module:
            # Reset
            self.process_module = False

    def visit_importfrom(self, node):
        """Triggered when a from statement is seen."""
        if self.process_module:
            if not node.modname.startswith("tests.integration"):
                return
            # Store salt imported modules
            for module, import_as in node.names:
                if import_as:
                    self._check_moved_imports(node, module, import_as)
                    continue
                self._check_moved_imports(node, module)

    def visit_classdef(self, node):
        for base in node.bases:
            if not hasattr(base, "attrname"):
                continue
            if base.attrname in ("TestCase",):
                msg = f"Please use 'from tests.support.unit import {base.attrname}'"
                self.add_message("moved-test-case-class", node=node, args=(msg,))
            if base.attrname in ("ModuleCase", "SyndicCase", "ShellCase", "SSHCase"):
                msg = f"Please use 'from tests.support.case import {base.attrname}'"
                self.add_message("moved-test-case-class", node=node, args=(msg,))
            if base.attrname in (
                "AdaptedConfigurationTestCaseMixin",
                "ShellCaseCommonTestsMixin",
                "SaltMinionEventAssertsMixin",
            ):
                msg = f"Please use 'from tests.support.mixins import {base.attrname}'"
                self.add_message("moved-test-case-mixin", node=node, args=(msg,))

    def _check_moved_imports(self, node, module, import_as=None):
        for name, name_as in node.names:
            if name not in ("ModuleCase", "SyndicCase", "ShellCase", "SSHCase"):
                continue
            if name_as:
                msg = f"Please use 'from tests.support.case import {name} as {name_as}'"
            else:
                msg = f"Please use 'from tests.support.case import {name}'"
            self.add_message("moved-test-case-class", node=node, args=(msg,))


BLACKLISTED_FUNCTIONS_MSGS = {
    "E9601": (
        "Use of blacklisted function %s (use %s instead)",
        "blacklisted-function",
        "Used a function marked as blacklisted",
    ),
}


class BlacklistedFunctionsChecker(BaseChecker):
    name = "blacklisted-functions"
    msgs = BLACKLISTED_FUNCTIONS_MSGS
    priority = -2
    max_depth = 20

    options = (
        (
            "blacklisted-functions",
            {
                "default": "",
                "type": "string",
                "metavar": "bad1=good1,bad2=good2",
                "help": "List of blacklisted functions and their recommended replacements",
            },
        ),
    )

    def open(self):
        self.blacklisted_functions = {}
        blacklist = [x.strip() for x in self.linter.config.blacklisted_functions.split(",")]
        for item in blacklist:
            try:
                key, val = (x.strip() for x in item.split("="))
            except ValueError:
                pass
            else:
                self.blacklisted_functions[key] = val

    def _get_full_name(self, node):
        try:
            func = utils.safe_infer(node.func)
            if func.name.__str__() == "Uninferable":
                return None
        except Exception:
            func = None
        if func is None:
            return None
        ret = []
        depth = 0
        while func is not None:
            depth += 1
            if depth > self.max_depth:
                # Prevent endless loop
                return None
            try:
                ret.append(func.name)
            except AttributeError:
                return None
            func = func.parent
        # ret will contain the levels of the function from last to first (e.g.
        # ['walk', 'os']. Reverse it and join with dots to get the correct
        # full name for the function.
        return ".".join(ret[::-1])

    def visit_call(self, node):
        if self.blacklisted_functions:
            full_name = self._get_full_name(node)
            if full_name is not None:
                with contextlib.suppress(KeyError):
                    self.add_message(
                        "blacklisted-function",
                        node=node,
                        args=(full_name, self.blacklisted_functions[full_name]),
                    )


def register(linter):
    """Required method to auto register this checker."""
    linter.register_checker(ResourceLeakageChecker(linter))
    linter.register_checker(BlacklistedImportsChecker(linter))
    linter.register_checker(MovedTestCaseClassChecker(linter))
    linter.register_checker(BlacklistedLoaderModulesUsageChecker(linter))
    linter.register_checker(BlacklistedFunctionsChecker(linter))
